! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      
      implicit none
      
      integer :: time0, time1, clock_rate

      
      contains

      
      subroutine how_many_extra_profile_header_items( &
            id, id_extra, num_extra_header_items)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: num_extra_header_items
         num_extra_header_items = 1
      end subroutine how_many_extra_profile_header_items
   
   
      subroutine data_for_extra_profile_header_items( &
            id, id_extra, num_extra_header_items, &
            extra_header_item_names, extra_header_item_vals, ierr)
         integer, intent(in) :: id, id_extra, num_extra_header_items
         character (len=*), pointer :: extra_header_item_names(:)
         real(dp), pointer :: extra_header_item_vals(:)
         integer, intent(out) :: ierr
         ierr = 0
         extra_header_item_names(1) = 'profileX'
         extra_header_item_vals(1) = 3.14159d0
      end subroutine data_for_extra_profile_header_items

      
      subroutine how_many_extra_history_header_items( &
            id, id_extra, num_extra_header_items)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: num_extra_header_items
         num_extra_header_items = 1
      end subroutine how_many_extra_history_header_items
   
   
      subroutine data_for_extra_history_header_items( &
            id, id_extra, num_extra_header_items, &
            extra_header_item_names, extra_header_item_vals, ierr)
         integer, intent(in) :: id, id_extra, num_extra_header_items
         character (len=*), pointer :: extra_header_item_names(:)
         real(dp), pointer :: extra_header_item_vals(:)
         integer, intent(out) :: ierr
         ierr = 0
         extra_header_item_names(1) = 'historyX'
         extra_header_item_vals(1) = 123456.789d0
      end subroutine data_for_extra_history_header_items
      
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         s% how_many_extra_profile_header_items => &
            how_many_extra_profile_header_items
         s% data_for_extra_profile_header_items => &
            data_for_extra_profile_header_items
         
         s% how_many_extra_history_header_items => &
            how_many_extra_history_header_items
         s% data_for_extra_history_header_items => &
            data_for_extra_history_header_items
         
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  
         
      end subroutine extras_controls
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         call system_clock(time0,clock_rate)
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      
      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         real(dp) :: dt
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         call system_clock(time1,clock_rate)
         dt = dble(time1 - time0) / clock_rate / 60
         write(*,'(/,a50,f12.2,99i10/)') 'runtime (minutes), retries, backups, steps', &
            dt, s% num_retries, s% num_backups, s% model_number
         ierr = 0
      end subroutine extras_after_evolve
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

! set how many reaction eps_nuc we want in the profiles
         how_many_extra_profile_columns = 2


      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         use net_lib, only: net_work_size
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k, op_err, net_lwork
         logical :: okay
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         net_lwork = net_work_size(s% net_handle, ierr)        


! set the profile column names 1 to how_many_extra_profile_column

         names(1) = 'eps_nuc_r34_pp2'
         names(2) = 'eps_nuc_r34_pp3'

         okay = .true.
!$OMP PARALLEL DO PRIVATE(k,op_err) 
         do k = 1, s% nz
            if (.not. okay) cycle
            op_err = 0
            call do1_net( &
               s, k, s% species, &
               s% num_reactions, net_lwork, &
               n, nz, vals, op_err)
            if (op_err /= 0) okay = .false.
         end do
!$OMP END PARALLEL DO        
      end subroutine data_for_extra_profile_columns


      subroutine do1_net( &
            s, k, species, num_reactions, net_lwork, &
            n, nz, vals, ierr)
         use rates_def, only: std_reaction_Qs, std_reaction_neuQs, i_rate
         use net_def, only: Net_Info
         use net_lib, only: net_get
         use chem_def, only: chem_isos, category_name
         use eos_def, only : i_eta
         use utils_lib,only: &
            is_bad_num, realloc_double, realloc_double3
         type (star_info), pointer :: s         
         integer, intent(in) :: k, species, num_reactions, net_lwork, n, nz
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
 
         integer :: i, j, screening_mode
         real(dp) :: log10_rho, log10_T, alfa, beta, &
            d_eps_nuc_dRho, d_eps_nuc_dT, cat_factor
         real(dp), target :: net_work_ary(net_lwork)
         real(dp), pointer :: net_work(:)
         type (Net_Info), target :: net_info_target
         type (Net_Info), pointer :: netinfo
         
         character (len=100) :: message
         real(dp), pointer :: reaction_neuQs(:)
         integer :: sz
         real(dp) :: eps_nuc_factor
         
         logical, parameter :: dbg = .false.

         include 'formats'
         
         ierr = 0
         
         net_work => net_work_ary
         netinfo => net_info_target
      
         log10_rho = s% lnd(k)/ln10
         log10_T = s% lnT(k)/ln10

         screening_mode = get_screening_mode(s,ierr)         
         if (ierr /= 0) then
            write(*,*) 'unknown string for screening_mode: ' // trim(s% screening_mode)
            stop 'do1_net'
            return
         end if
      
         call net_get( &
            s% net_handle, .false., netinfo, species, num_reactions, s% xa(1:species,k), &
            s% T(k), log10_T, s% rho(k), log10_Rho, &
            s% abar(k), s% zbar(k), s% z2bar(k), s% ye(k), &
            s% eta(k), s% d_eos_dlnd(i_eta,k), s% d_eos_dlnT(i_eta,k), &
            s% rate_factors, s% weak_rate_factor, &
            std_reaction_Qs, std_reaction_neuQs, .false., .false., &
            s% eps_nuc(k), d_eps_nuc_dRho, d_eps_nuc_dT, s% d_epsnuc_dx(:,k), & 
            s% dxdt_nuc(:,k), s% d_dxdt_nuc_dRho(:,k), s% d_dxdt_nuc_dT(:,k), s% d_dxdt_nuc_dx(:,:,k), &
            screening_mode, s% theta_e(k), &
            s% eps_nuc_categories(:,k), &
            s% eps_nuc_neu_total(k), net_lwork, net_work, ierr)
      
         if (ierr /= 0) then
            write(*,*) 'do1_net: net_get failure for cell ', k
            return
         end if
         
         call show_stuff(s,k,net_lwork,net_work,n,nz,vals)

      end subroutine do1_net
      
      
      integer function get_screening_mode(s,ierr)
         use rates_lib, only: screening_option
         type (star_info), pointer :: s 
         integer, intent(out) :: ierr
         include 'formats'
         ierr = 0
         if (s% screening_mode_value >= 0) then
            get_screening_mode = s% screening_mode_value
            return
         end if
         get_screening_mode = screening_option(s% screening_mode, ierr)
         if (ierr /= 0) return
         s% screening_mode_value = get_screening_mode
         !write(*,2) 'get_screening_mode ' // &
         !   trim(s% screening_mode), get_screening_mode
      end function get_screening_mode
      
      
      
      subroutine show_stuff(s,k,lwork,work,n,nz,vals)
         use chem_def
         use rates_def
         use rates_lib, only: rates_reaction_id
         use net_lib, only: get_reaction_id_table_ptr, get_net_rate_ptrs
         use crlibm_lib, only: log10_cr
         type (star_info), pointer :: s         
         integer, intent(in) :: k, lwork, n, nz
         real(dp), pointer :: work(:)
         real(dp) :: vals(nz,n)

         integer, pointer :: reaction_id(:) ! maps net reaction number to reaction id
         integer :: i, j, ierr, species, num_reactions, rate_id
         real(dp), pointer, dimension(:) :: &
            rate_screened, rate_screened_dT, rate_screened_dRho, &
            rate_raw, rate_raw_dT, rate_raw_dRho

         include 'formats'
         
         ierr = 0
         num_reactions = s% num_reactions
         
         call get_net_rate_ptrs(s% net_handle, &
            rate_screened, rate_screened_dT, rate_screened_dRho, &
            rate_raw, rate_raw_dT, rate_raw_dRho, lwork, work, &
            ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in get_net_rate_ptrs'
            stop 1
         end if
         
         call get_reaction_id_table_ptr(s% net_handle, reaction_id, ierr) 
         if (ierr /= 0) return

! set the character string name of the reaction rates
! and fill vals(k,how_many_extra_profile_columns)

! pp2         
         rate_id = rates_reaction_id('r34_pp2')
         if (rate_id <= 0) then
            write(*,*) 'failed to find reaction rate id -- not valid name?'
            vals(k,1) = 0
            return
         end if
         vals(k,1) = get_rate(rate_id)
         
! pp3
         rate_id = rates_reaction_id('r34_pp3')
         if (rate_id <= 0) then
            write(*,*) 'failed to find reaction rate id -- not valid name?'
            vals(k,2) = 0
            return
         end if
         vals(k,2) = get_rate(rate_id)
         
         
         contains
         
         real(dp) function get_rate(id)
            integer, intent(in) :: id
            integer :: j
            include 'formats'

            get_rate = -99
            do j=1,num_reactions

               if (reaction_id(j) /= rate_id) cycle

!                write(6,'(a,i4,i4,1p2e12.4)') 'found it', id, size(std_reaction_Qs), & 
!                           std_reaction_Qs(ir34_pp2), std_reaction_Qs(ir34_pp3)

               if (rate_screened(j) < 1d-20) then
                  get_rate = -99

! rates are in num_reaction order, Q's are in irate order
! eps_nuc from a specific reaction

               else
                  get_rate = log10_cr(rate_screened(j) * & 
                             (std_reaction_Qs(rate_id) - std_reaction_neuQs(rate_id)) * Qconv)
               end if
               return
            end do

            write(*,*) 'failed to find reaction rate id -- not in current net?'
            get_rate = -99

         end function get_rate
         
      end subroutine show_stuff

      

      ! returns either keep_going or terminate.
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)
      end function extras_finish_step
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
